package net.sf.logdistiller;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.Serializable;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.jdom.Element;

/**
 * One condition for a LogEvent to be selected in a group. It consists of matches on attributes of the LogEvent.
 */
public class Condition
    implements Serializable
{
    private static final long serialVersionUID = 9085774334169039381L;

    private final Match[] matches;

    public static final Condition ACCEPT_ALL = new Condition( null, new Match[0] );

    /**
     * @since 1.1
     */
    private final Set tags;

    public Condition( String tags, Match[] matches )
    {
        if ( tags != null )
        {
            this.tags = new HashSet();
            String[] values = tags.split( "," );
            for ( int i = values.length - 1; i > 0; i-- )
            {
                String tag = values[i].trim();
                if ( tag.length() > 0 )
                {
                    this.tags.add( tag );
                }
            }
        }
        else
        {
            this.tags = null;
        }
        this.matches = (Match[]) ArrayUtils.clone( matches );
    }

    public boolean match( LogEvent le )
    {
        int count = matches.length;
        for ( int i = 0; i < count; i++ )
        {
            if ( !matches[i].match( le ) )
            {
                return false;
            }
        }
        return true;
    }

    /**
     * @since 1.1
     * @param tag tag value looked for
     * @return <code>true</code> if this condition was flagged with corresponding tag
     */
    public boolean hasTag( String tag )
    {
        return ( tags == null ) ? false : tags.contains( tag );
    }

    public Element dump()
    {
        Element elmt = new Element( "condition" );
        if ( tags != null )
        {
            elmt.setAttribute( "tags", StringUtils.join( tags, ", " ) );
        };
        int count = ( matches == null ) ? 0 : matches.length;
        for ( int i = 0; i < count; i++ )
        {
            elmt.addContent( matches[i].dump() );
        }
        return elmt;
    }
}
