package net.sf.logdistiller.gui;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Frame;
import java.awt.LayoutManager;
import java.awt.event.ActionEvent;
import java.io.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.swing.*;

import com.jgoodies.forms.factories.ButtonBarFactory;
import com.jgoodies.forms.factories.DefaultComponentFactory;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;

import net.sf.logdistiller.LogType;
import net.sf.logdistiller.LogTypes;
import net.sf.logdistiller.util.FormatUtil;
import net.sf.logdistiller.util.PropertiesReplacer;

/**
 * Dialog to create a new rules configuration file: you select the log type and give some basic information (an id, STMP
 * server, url), and it generates a new rules configuration. Then you just need to refine the rules.
 */
public class NewDialog
{
    private File currentDir;

    public NewDialog( File directory )
    {
        currentDir = directory;
    }

    protected JPanel panel = new JPanel();

    private JTextField tfId = new JTextField( "rules-id", 15 );

    private JTextField tfDescription = new JTextField( "rules description" );

    private JComboBox cbLogtype;

    private JTextField tfSmtp = new JTextField( "smtp.provider.com" );

    private JTextField tfMailTo = new JTextField( "admin@localhost" );

    private JTextField tfHttp = new JTextField( "http://mysite.com/logdistiller/logs" );

    private JButton btSaveAs = new JButton( "save as..." );

    private JButton btCancel = new JButton( "cancel" );

    public static JComboBox createCbLogtype( boolean withUndefined )
    {
        List logtypes = LogTypes.getAllLogTypes();
        int len = logtypes.size();
        String[] logtypeIds = new String[len];
        for ( int i = 0; i < len; i++ )
        {
            logtypeIds[i] = ( (LogType) logtypes.get( i ) ).getId();
        }
        JComboBox cb = new JComboBox( logtypeIds );
        if ( withUndefined )
        {
            cb.insertItemAt( "undefined", 0 );
            cb.setSelectedIndex( 0 );
        }
        return cb;
    }

    private void initComponents()
    {
        cbLogtype = createCbLogtype( false );

        btSaveAs.addActionListener( SwingAdapter.getActionListener( this, "btSaveAsActionPerformed" ) );
        btCancel.addActionListener( SwingAdapter.getActionListener( this, "btCancelActionPerformed" ) );
    }

    private void addLabel( String label, String id )
    {
        panel.add( new JLabel( label ), id );
    }

    private void addSeparator( String label, String id )
    {
        panel.add( DefaultComponentFactory.getInstance().createSeparator( label ), id );
    }

    public LayoutManager createLayout( String layoutId )
    {
        return MainPanel.SIMPLE_LAYOUT_CONSTRAINTS_MANAGER.getLayout( layoutId );
    }

    public JComponent buildPanel()
    {
        initComponents();

        panel.setBorder( com.jgoodies.forms.factories.Borders.DIALOG_BORDER );
        panel.setLayout( createLayout( "newDialog" ) );

        addSeparator( "main parameters", "sepMain" );
        addLabel( "configuration id:", "lblId" );
        panel.add( tfId, "tfId" );
        addLabel( "configuration description:", "lblDescription" );
        panel.add( tfDescription, "tfDescription" );
        addLabel( "logtype:", "lblLogtype" );
        panel.add( cbLogtype, "cbLogtype" );

        addSeparator( "advanced parameters", "sepOptions" );
        addLabel( "SMTP host:", "lblSmtp" );
        panel.add( tfSmtp, "tfSmtp" );
        addLabel( "mail reports to:", "lblMailTo" );
        panel.add( tfMailTo, "tfMailTo" );
        addLabel( "reports publication URL:", "lblHttp" );
        panel.add( tfHttp, "tfHttp" );

        Component bbar = ButtonBarFactory.buildRightAlignedBar( new JButton[] { btSaveAs, btCancel } );
        panel.add( bbar, "bbar" );

        return panel;
    }

    private JDialog createDialog( Component parent )
    {
        Frame frame =
            parent instanceof Frame ? (Frame) parent : (Frame) SwingUtilities.getAncestorOfClass( Frame.class, parent );
        JDialog dialog = new JDialog( frame, "new classification rules configuration", true );

        Container contentPane = dialog.getContentPane();
        contentPane.setLayout( new BorderLayout() );
        contentPane.add( buildPanel(), BorderLayout.CENTER );

        dialog.pack();
        dialog.setLocationRelativeTo( parent );
        return dialog;
    }

    private JDialog dialog;

    public File showNewConfigurationDialog( Component parent )
    {
        dialog = createDialog( parent );
        dialog.show();
        dialog.dispose();
        dialog = null;
        return configurationFile;
    }

    private File configurationFile;

    public void btSaveAsActionPerformed( ActionEvent ae )
    {
        JFileChooser fc = new JFileChooser( currentDir );
        fc.setFileFilter( new XmlFileFilter() );
        fc.setSelectedFile( new File( currentDir, tfId.getText() + ".xml" ) );
        int ret = fc.showSaveDialog( dialog );
        if ( ret == JFileChooser.APPROVE_OPTION )
        {
            saveConfigurationFile( fc.getSelectedFile() );
            configurationFile = fc.getSelectedFile();
            if ( dialog != null )
            {
                dialog.setVisible( false );
            }
        }
    }

    /**
     * Saves a new rules configuration to a file.
     *
     * @param fileTo the file to save to
     */
    private void saveConfigurationFile( File fileTo )
    {
        try
        {
            PrintWriter out = new PrintWriter( new OutputStreamWriter( new FileOutputStream( fileTo ), "utf-8" ) );
            generateConfiguration( out );
            out.close();
        }
        catch ( FileNotFoundException fnfe )
        {
            fnfe.printStackTrace();
        }
        catch ( UnsupportedEncodingException uee )
        {
            uee.printStackTrace();
        }
    }

    /**
     * Generate a new configuration stream by filling <code>reference-rules.xml</code> with values taken from the GUI.
     *
     * @param out the stream to write to (in UTF-8 encoding)
     */
    private void generateConfiguration( PrintWriter out )
    {
        LogType logtype = LogTypes.getLogType( (String) cbLogtype.getSelectedItem() );
        LogType.Description description = logtype.newDescription( new HashMap() );

        String logeventAttributes = FormatUtil.join( ",", description.getAttributeNames() );
        int dateAttributeIndex = description.getTimestampAttribute();

        try
        {
            String reference = IOUtils.toString( MainPanel.class.getResourceAsStream( "reference-rules.xml" ), "UTF-8" );

            Map properties = new HashMap();
            properties.put( "id", tfId.getText() );
            properties.put( "description", tfDescription.getText() );
            properties.put( "logtype", cbLogtype.getSelectedItem() );
            properties.put( "logtype.attributes", logeventAttributes );
            properties.put( "mail.to", tfMailTo.getText() );
            properties.put( "mail.smtp.host", tfSmtp.getText() );
            properties.put( "logs.url", tfHttp.getText() );
            properties.put( "logtype.groups", description.getDefaultSpecificGroups() );

            String skipCondition = // default matching condition for skip group, if no timestamp available
                "      <!-- this logtype does not have a timestamp: we can't really skip events with bad date -->\n"
                    + "      <match attribute=\"logSource\" type=\"equals\"></match>\n"
                    + "      <!-- if this logtype had a timestamp, the condition would look like the following: -->\n"
                    + "      <!--match attribute=\"(timestamp attribute?)\" type=\"!contains\">${date}</match!-->";
            if ( dateAttributeIndex >= 0 )
            {
                skipCondition =
                    "      <match attribute=\"" + description.getAttributeName( dateAttributeIndex )
                        + "\" type=\"!contains\">${date}</match>";
            }
            properties.put( "skip.condition", skipCondition );

            // let's compute the sampling.rule property for this LogType
            String samplingAttributes = description.getDefaultSamplingAttributes();
            String samplingRule = // default comment when no sampling attributes defined
                "\n    <!--plugin type=\"sampling\">\n"
                + "      <param name=\"attributes\">most discriminating attribute</param>\n"
                + "    </plugin-->";
            if ( samplingAttributes.length() > 0 )
            {
                // ok, there are sampling attributes
                samplingRule =
                    "\n    <plugin type=\"sampling\">\n"
                    + "      <param name=\"attributes\">" + samplingAttributes + "</param>\n"
                    + "    </plugin>";
            }
            properties.put( "sampling.rule", samplingRule );

            PropertiesReplacer repl = new PropertiesReplacer( properties, "#(", ")" );

            out.print( repl.replaceProperties( reference ) );
        }
        catch ( IOException ioe )
        {
            ioe.printStackTrace();
        }
    }

    public void btCancelActionPerformed( ActionEvent ae )
    {
        configurationFile = null;
        if ( dialog != null )
        {
            dialog.setVisible( false );
        }
    }
}
