package net.sf.logdistiller.logtypes;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.*;
import java.text.ParseException;

import net.sf.logdistiller.LogEvent;
import net.sf.logdistiller.LogType;

/**
 * Generic log event for simple files where one line is one logevent, with an unique attribute named <code>text</code>.
 * By default, the classification rules generated by the GUI for this type of logs won't be able to sort events, as the
 * log event structure is too generic.
 */
public class SimpleLineLogEvent
    extends LogEvent
{
    public final static String ID = "simple";

    public final String text;

    public final static LogType LOGTYPE = new LogType.Basic( ID );

    private final static String[] ATTRIBUTE_NAMES = { "logSource", "text" };

    public final static LogType.Description DESCRIPTION = new Description( (LogType.Basic) LOGTYPE, ATTRIBUTE_NAMES );

    public SimpleLineLogEvent( Factory factory, String rawLog )
        throws ParseException
    {
        super( factory, rawLog );
        this.text = rawLog;
        setAttributes( new String[] { factory.getLogSource(), text } );
    }

    private static class Description
        extends LogType.Description
    {
        public Description( LogType.Basic logtype, String[] attributeNames )
        {
            super( logtype, attributeNames );
            logtype.setDescription( this );
        }

        public LogEvent.Factory newFactory( Reader reader, String logSource )
            throws IOException
        {
            return new Factory( this, reader, logSource );
        }

        public int getTimestampAttribute()
        {
            return -1;
        }
    }

    private static class Factory
        extends LogEvent.Factory
    {
        private final LineNumberReader reader;

        public Factory( Description description, Reader reader, String logSource )
            throws FileNotFoundException
        {
            super( description, logSource );
            this.reader = new LineNumberReader( reader );
        }

        protected LogEvent readNextEvent()
            throws IOException, ParseException
        {
            String curLine = reader.readLine();
            if ( curLine == null )
            {
                // EOF
                return null;
            }
            return new SimpleLineLogEvent( this, curLine );
        }
    }
}
