package net.sf.logdistiller.publishers;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.*;
import java.util.*;

import com.sun.syndication.feed.synd.*;
import com.sun.syndication.io.FeedException;
import com.sun.syndication.io.SyndFeedInput;
import com.sun.syndication.io.SyndFeedOutput;

import net.sf.logdistiller.*;

/**
 * Adds LogDistillation reports to a feed file (at the beginning). Any feed format supported by Rome can be used.
 * <p>
 * Parameters (as direct publisher param or global <code>feed.*</code> param):
 * <ul>
 * <li><code>filename</code> (default: <code>feed.xml</code>): the feed file name (relative to output directory)</li>
 * <li><code>maxEntries</code> (default: 100): maximum entries in the feed file: if new entries are added, oldest ones
 * are removed</li>
 * </ul>
 * <p>
 * When the feed file must be created, these parameters are used:
 * <ul>
 * <li><code>type</code> (default: <code>atom_0.3</code>): the feed type (any value supported by Rome)</li>
 * <li><code>title</code>: the feed title (default: "LogDistiller [logdistillation id]"</li>
 * <li><code>link</code>: the feed link (default: LogDistiller's site)</li>
 * <li><code>description</code>: the feed description (default: "LogDistiller reports for [logdistillation id]"</li>
 * </ul>
 *
 * @since 0.7
 */
public class FeedPublisher
    extends Publisher
{
    public final static String DEFAULT_FEED_TYPE = "atom_0.3";

    public String getId()
    {
        return "feed";
    }

    public void publish( LogDistillation logdistillation, LogDistiller.Report report )
        throws IOException, PublishException
    {
        // prepare the data for the entry
        String content = logdistillation.getContent();
        String subject =
            "[logdistiller:" + logdistillation.getDefinition().getId() + "]"
                + ( ( content == null ) ? "" : ( " " + content ) ) + " global report: "
                + logdistillation.getEventCount() + " logevents";
        ReportFormat format = ReportFormats.getReportFormat( report.getFormat() );
        StringWriter sw = new StringWriter();
        format.report( logdistillation, sw );
        String body = sw.toString();
        String url = logdistillation.getDefinition().getOutput().getUrl();

        // create the new entry
        SyndEntry entry = createEntry( subject, url, format.getContentType(), body );

        // add it to the feed (creating it if necessary)
        addEntry( entry, logdistillation, report );
    }

    public void publish( LogDistillation.Group group, LogDistiller.Report report )
        throws IOException, PublishException
    {
        // prepare the data for the entry
        LogDistillation logdistillation = group.getLogdistillation();
        String content = logdistillation.getContent();
        String subject =
            "[logdistiller:" + logdistillation.getDefinition().getOutput().getId() + "]"
                + ( ( content == null ) ? "" : ( " " + content ) ) + " report for group "
                + group.getDefinition().getId() + ": " + group.getEventCount() + " logevents";
        ReportFormat format = ReportFormats.getReportFormat( report.getFormat() );
        StringWriter sw = new StringWriter();
        format.report( group, sw );
        String body = sw.toString();
        String url = logdistillation.getDefinition().getOutput().getUrl();

        // create the new entry
        SyndEntry entry = createEntry( subject, url, format.getContentType(), body );

        // add it to the feed (creating it if necessary)
        addEntry( entry, logdistillation, report );
    }

    /**
     * Create a new feed for the corresponding report.
     *
     * @param logdistillation
     * @param report
     * @return the new feed
     */
    private SyndFeed createFeed( LogDistillation logdistillation, LogDistiller.Report report )
    {
        String type = report.getParam( "type", DEFAULT_FEED_TYPE );
        String title = report.getParam( "title", "Logdistiller " + logdistillation.getDefinition().getOutput().getId() );
        String url = logdistillation.getDefinition().getOutput().getUrl();
        String link = report.getParam( "link", ( url == null ) ? "http://logdistiller.sf.net/" : url );
        String description =
            report.getParam( "description", "LogDistiller reports for "
                + logdistillation.getDefinition().getOutput().getId() );

        SyndFeed feed = new SyndFeedImpl();
        feed.setFeedType( type );
        feed.setTitle( title );
        feed.setLink( link );
        feed.setDescription( description );

        return feed;
    }

    /**
     * Create a feed entry.
     *
     * @param title the title of the entry
     * @param link the link of the entry (or <code>null</code>)
     * @param contentType the content-type of the entry
     * @param contentValue the content of the entry
     * @return a new feed entry
     */
    private SyndEntry createEntry( String title, String link, String contentType, String contentValue )
    {
        SyndEntry entry = new SyndEntryImpl();
        entry.setTitle( title );
        if ( link != null )
        {
            entry.setLink( link );
        }
        entry.setPublishedDate( new Date() );

        SyndContent content = new SyndContentImpl();
        content.setType( contentType );
        content.setValue( contentValue );
        List contents = new ArrayList();
        contents.add( content );

        entry.setContents( contents );
        return entry;
    }

    /**
     * Add an entry to a feed report, creating the feed if necessary.
     *
     * @param entry the feed entry
     * @param logdistillation the current logdistillation
     * @param report the report configuration
     * @throws IOException
     * @throws PublishException
     */
    private void addEntry( SyndEntry entry, LogDistillation logdistillation, LogDistiller.Report report )
        throws IOException, PublishException
    {
        String filename = report.getParam( "filename", "feed.xml" );
        int maxEntries = Integer.parseInt( report.getParam( "maxEntries", "100" ) );
        File file = logdistillation.newDestinationFile( filename );

        // read or create the feed
        SyndFeed feed;
        if ( file.canRead() && ( file.length() > 0 ) )
        {
            SyndFeedInput input = new SyndFeedInput();
            try
            {
                feed = input.build( file );
            }
            catch ( FeedException fe )
            {
                throw new PublishException( "error while reading feed " + filename, fe );
            }
        }
        else
        {
            feed = createFeed( logdistillation, report );
        }

        // add the new entry
        List entries = feed.getEntries();
        entries.add( 0, entry );
        if ( entries.size() > maxEntries )
        {
            entries.remove( entries.size() - 1 );
        }
        feed.setEntries( entries );

        // write the feed content to the disk
        SyndFeedOutput output = new SyndFeedOutput();
        // write to a temporary file that will be swapped with the current file only on success
        File tmpFile = logdistillation.newDestinationFile( filename + ".tmp" );
        try
        {
            output.output( feed, tmpFile );
            // ok, we were able to write the whole feed: now we can rotate with the final file
            file.delete();
            tmpFile.renameTo( file );
        }
        catch ( FeedException fe )
        {
            throw new PublishException( "error while writing temporary feed " + filename + ".tmp", fe );
        }
        finally
        {
            tmpFile.delete();
        }
    }
}
