package net.sf.logdistiller.publishers;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.*;
import java.util.Properties;
import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.activation.FileDataSource;
import javax.mail.*;
import javax.mail.internet.*;

import net.sf.logdistiller.*;

/**
 * Send mail report of a LogDistillation. When the report is a group report, a file containing logevents (if saved) is
 * added as attachment. Global output parameters are used as JavaMail session properties: the most important property is
 * <code>mail.smtp.host</code>, but see JavaMail documentation to find all available properties.
 * <p>
 * Parameters:
 * <ul>
 * <li><code>to</code>, <code>cc</code> (at least one): recipients of the mail, comma separated values if multiple
 * recipients
 * <li><code>from</code> (default: <code>smtp.from</code> global output parameter, or
 * <code>logdistiller@sourceforge.net</code>
 * <li><code>attachLogs</code> (default: <code>true</code>): add corresponding logevents as attachment
 * </ul>
 */
public class MailPublisher
    extends Publisher
{
    public String getId()
    {
        return "mail";
    }

    public void publish( LogDistillation logdistillation, LogDistiller.Report report )
        throws IOException, PublishException
    {
        String content = logdistillation.getContent();
        String subject =
            "[logdistiller:" + logdistillation.getDefinition().getOutput().getId() + "]"
                + ( ( content == null ) ? "" : ( " " + content ) ) + " global report: "
                + logdistillation.getEventCount() + " logevents";
        ReportFormat format = ReportFormats.getReportFormat( report.getFormat() );
        StringWriter sw = new StringWriter();
        format.report( logdistillation, sw );
        String body = sw.toString();
        try
        {
            sendMail( logdistillation, report, subject, body, format.getContentType(), null );
        }
        catch ( MessagingException me )
        {
            throw new PublishException( "error while sending global report by mail: " + me.getMessage(), me );
        }
    }

    public void publish( LogDistillation.Group group, LogDistiller.Report report )
        throws IOException, PublishException
    {
        LogDistillation logdistillation = group.getLogdistillation();
        String content = logdistillation.getContent();
        String subject =
            "[logdistiller:" + logdistillation.getDefinition().getOutput().getId() + "]"
                + ( ( content == null ) ? "" : ( " " + content ) ) + " report for group "
                + group.getDefinition().getId() + ": " + group.getEventCount() + " logevents";
        ReportFormat format = ReportFormats.getReportFormat( report.getFormat() );
        StringWriter sw = new StringWriter();
        format.report( group, sw );
        String body = sw.toString();
        try
        {
            sendMail( logdistillation, report, subject, body, format.getContentType(), group.getLogFile() );
        }
        catch ( MessagingException me )
        {
            throw new PublishException( "error while sending group report by mail: " + me.getMessage(), me );
        }
    }

    protected void sendMail( LogDistillation logdistillation, LogDistiller.Report report, String subject,
                             String content, String mimetype, File attach )
        throws MessagingException
    {
        LogDistiller ld = logdistillation.getDefinition();
        String to = report.getParam( "to" );
        String cc = report.getParam( "cc" );
        if ( ( to == null ) && ( cc == null ) )
        {
            // nobody to send to: TODO add warning
            return;
        }
        Properties props = new Properties();
        props.putAll( ld.getOutput().getParams() );
        Session session = Session.getDefaultInstance( props, null );
        MimeMessage msg = new MimeMessage( session );

        msg.setSubject( subject );
        Multipart mp = new MimeMultipart();
        BodyPart body = new MimeBodyPart();
        body.setContent( content, mimetype );
        mp.addBodyPart( body );
        msg.setContent( mp );
        if ( to != null )
        {
            msg.setRecipients( Message.RecipientType.TO, InternetAddress.parse( to ) );
        }
        if ( cc != null )
        {
            msg.setRecipients( Message.RecipientType.CC, InternetAddress.parse( cc ) );
        }
        String from = report.getParam( "from", "logdistiller@sourceforge.net" );
        msg.setFrom( new InternetAddress( from ) );

        String attachLogs = report.getParam( "attachLogs", "true" );
        if ( ( attach != null ) && ( attach.exists() ) && ( "true".equals( attachLogs ) ) )
        {
            BodyPart attachment = new MimeBodyPart();
            DataSource ds = new FileDataSource( attach );
            attachment.setDataHandler( new DataHandler( ds ) );
            attachment.setFileName( attach.getName() );
            mp.addBodyPart( attachment );
        }

        sendMail( msg );
    }

    protected void sendMail( Message msg )
        throws MessagingException
    {
        Transport.send( msg );
    }
}
