package net.sf.logdistiller.reports;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;
import java.text.DateFormat;
import java.util.Date;
import java.util.Iterator;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.SystemUtils;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import net.sf.logdistiller.LogDistillation;
import net.sf.logdistiller.LogDistiller;
import net.sf.logdistiller.ReportFormat;
import net.sf.logdistiller.util.FormatUtil;

/**
 * Pure text report format.
 */
public class TextReport
    extends ReportFormat
{
    private final static DateFormat DF = DateFormat.getDateTimeInstance( DateFormat.FULL, DateFormat.LONG );

    public TextReport()
    {
        super( "txt" );
    }

    public String getContentType()
    {
        return "text/plain";
    }

    public String getFileExtension()
    {
        return "txt";
    }

    public void report( LogDistillation ld, Writer output )
        throws IOException
    {
        LogDistiller definition = ld.getDefinition();
        PrintWriter out = new PrintWriter( output );
        out.println( "LogDistillation result (LogDistiller " + ld.getVersion() + ")" );
        out.println();
        out.println( "LogDistiller used: " + definition.getOutput().getId() + " - " + definition.getDescription() );
        out.println( "log content processed: " + ( ( ld.getContent() == null ) ? "not specified" : ld.getContent() ) );
        out.println( "distillation started on " + DF.format( new Date( ld.getBeginTime() ) ) + ", finished on "
            + DF.format( new Date( ld.getEndTime() ) ) + ": "
            + FormatUtil.formatPeriod( ld.getEndTime() - ld.getBeginTime() ) );
        if ( ld.getSkippedEventCount() > 0 )
        {
            out.println( "skipped " + ld.getSkippedEventCount() + " events ("
                + FormatUtil.formatSize( ld.getSkippedEventBytes() ) + ")" );
        }
        String logsUrl = definition.getOutput().getUrl();
        if ( logsUrl != null )
        {
            out.println( "reports and logs url: " + logsUrl );
        }
        Iterator iter = definition.getWarnings().iterator();
        while ( iter.hasNext() )
        {
            out.println( "WARNING: " + iter.next() );
        }
        out.println();
        if ( ld.getEventCount() == 0 )
        {
            out.println( "found 0 event to classify." );
        }
        else
        {
            out.println( "found " + ld.getEventCount() + " events (" + FormatUtil.formatSize( ld.getEventBytes() )
                + ") classified as follow:" );
            // report for categories, and corresponding groups
            Iterator categoriesIterator = ld.listCategoriesToReport().iterator();
            while ( categoriesIterator.hasNext() )
            {
                LogDistillation.Category category = (LogDistillation.Category) categoriesIterator.next();
                int categoryEventCount = category.sumEventCount();
                LogDistiller.Category categoryDefinition = category.getDefinition();
                out.println( "*** " + categoryEventCount + " " + categoryDefinition.getId() + " ("
                    + FormatUtil.formatFraction( categoryEventCount, ld.getEventCount() ) + "): "
                    + categoryDefinition.getDescription() );
                int countLength = String.valueOf( categoryEventCount ).length();
                iter = category.listGroupsToReport().iterator();
                while ( iter.hasNext() )
                {
                    LogDistillation.Group group = (LogDistillation.Group) iter.next();
                    String fraction = formatFraction( group.getEventCount(), categoryEventCount );
                    out.println( "*" + fraction + ' ' + formatGroup( group, countLength, ld.getEventCount() ) );
                    addPluginsGlobalReport( group, new TextPluginReport( out, "*       ", group.getEventCount() ) );
                }
                out.println();
            }

            // report for groups without category
            int countLength = String.valueOf( ld.getEventCount() ).length();
            Iterator groupsIterator = ld.listGroupsToReport().iterator();
            while ( groupsIterator.hasNext() )
            {
                LogDistillation.Group group = (LogDistillation.Group) groupsIterator.next();
                out.println( formatGroup( group, countLength, ld.getEventCount() ) );
                addPluginsGlobalReport( group, new TextPluginReport( out, "   ", group.getEventCount() ) );
            }
        }
        out.flush();
    }

    private String formatGroup( LogDistillation.Group group, int countLength, int eventCount )
    {
        int groupEventCount = group.getEventCount();
        LogDistiller.Group groupDefinition = group.getDefinition();
        String sep = groupDefinition.continueProcessing() ? "+ " : "- ";
        return sep + StringUtils.leftPad( String.valueOf( groupEventCount ), countLength ) + " ("
            + formatFraction( groupEventCount, eventCount ) + ") " + groupDefinition.getId() + ": "
            + groupDefinition.getDescription();
    }

    private String formatFraction( long fraction, long total )
    {
        return StringUtils.leftPad( FormatUtil.formatFraction( fraction, total ), 6 );
    }

    public void report( LogDistillation.Group group, Writer output )
        throws IOException
    {
        LogDistiller.Group definition = group.getDefinition();
        PrintWriter out = new PrintWriter( output );
        out.println( "LogDistillation group result (LogDistiller " + LogDistiller.getVersion() + ")" );
        out.println();
        out.println( "LogDistiller group: " + definition.getId() + " - " + definition.getDescription() );
        out.println( "for LogDistiller: " + definition.getLogdistiller().getOutput().getId() + " - "
            + definition.getLogdistiller().getDescription() );
        out.println( "log content processed: "
            + ( ( group.getLogdistillation().getContent() == null ) ? "not specified"
                            : group.getLogdistillation().getContent() ) );
        out.println( "found " + group.getEventCount() + " events (" + FormatUtil.formatSize( group.getBytes() )
            + ") matching following definition:" );
        Element elmt = new Element( "conditions" );
        definition.dumpConditions( elmt );
        if ( elmt.getContentSize() > 0 )
        {
            out.print( "  " );
            new XMLOutputter( Format.getPrettyFormat().setLineSeparator( SystemUtils.LINE_SEPARATOR + "  " ) ).outputElementContent(
                                                                                                                                     elmt,
                                                                                                                                     output );
            out.println();
        }
        else
        {
            out.println( "  no condition specified: every log event accepted" );
        }

        String logsUrl = definition.getLogdistiller().getOutput().getUrl();
        if ( definition.getSave() )
        {
            out.println( "log events saved in file: " + group.getLogFile().getAbsolutePath() );
            out.print( "max save count: " );
            if ( group.getMaxSaveCount() < 0 )
            {
                out.println( "none" );
            }
            else
            {
                out.println( String.valueOf( group.getMaxSaveCount() ) );
            }
            out.print( "max save size: " );
            if ( group.getMaxSaveBytes() < 0 )
            {
                out.println( "none" );
            }
            else
            {
                out.println( String.valueOf( group.getMaxSaveBytes() / 1024 ) + " kb" );
            }
            if ( group.getSavedEventCount() < group.getEventCount() )
            {
                out.println( "save limit reached: " + group.getSavedEventCount() + " events" );
            }
            if ( logsUrl != null )
            {
                out.println( "log events url: "
                    + definition.getLogdistiller().getOutput().getUrl( group.getLogFile().getName() ) );
            }
        }
        else
        {
            out.println( "log events not saved" );
        }
        if ( logsUrl != null )
        {
            out.println( "reports and logs url: " + logsUrl );
        }
        if ( definition.continueProcessing() )
        {
            out.println( "continueProcessing: true" );
        }
        out.println();
        addPluginsGroupReport( group, new TextPluginReport( out, "", group.getEventCount() ) );
        out.flush();
    }

    private static class TextPluginReport
        extends PluginReport
    {
        private final PrintWriter out;

        private final String prefix;

        private final int groupEventCount;

        private final int countLength;

        public TextPluginReport( PrintWriter out, String prefix, int groupEventCount )
        {
            this.out = out;
            this.prefix = prefix;
            this.groupEventCount = groupEventCount;
            countLength = String.valueOf( groupEventCount ).length();
        }

        public void beginPluginReport( LogDistillation.Plugin plugin, String description )
        {
            out.println( prefix + plugin.getDefinition().getType() + ": " + description );
        }

        public void addLink( String filename, String description )
        {
            out.println( prefix + "file " + filename + ": " + description );
        }

        public void addParam( String param, String value )
        {
            out.println( prefix + param + ": " + value );
        }

        public void addItem( int count, String description )
        {
            out.println( prefix + "  # " + StringUtils.leftPad( String.valueOf( count ), countLength ) + " ("
                + StringUtils.leftPad( FormatUtil.formatFraction( count, groupEventCount ), 6 ) + ") " + description );
        }

        public void endPluginReport()
        {
            out.println( prefix.trim() );
        }
    }
}
